unit GuardMain;

interface

function Enter: Boolean;
procedure Leave;

implementation
uses
  Windows, Tools, ExeHdr, EventLogMessages;

{$region 'event log access'}
const
  EVENTLOG_SUCCESS          = $0000;
  EVENTLOG_ERROR_TYPE       = $0001;
  EVENTLOG_WARNING_TYPE     = $0002;
  EVENTLOG_INFORMATION_TYPE = $0004;
  EVENTLOG_AUDIT_SUCCESS    = $0008;
  EVENTLOG_AUDIT_FAILURE    = $0010;

type
  TEventLogger = class(TObject)
  private
    FName: String;
    FEventLog: Integer;
  public
    constructor Create(Name: String);
    destructor Destroy; override;
    procedure LogMessage(Message: String; EventType: DWord = 1;
      Category: Word = 0; ID: DWord = 0);
  end;

{ TEventLogger }

constructor TEventLogger.Create(Name: String);
begin
  FName := Name;
  FEventLog := 0;
end;

destructor TEventLogger.Destroy;
begin
  if FEventLog <> 0 then
    DeregisterEventSource(FEventLog);
  inherited Destroy;
end;

procedure TEventLogger.LogMessage(Message: String; EventType: DWord;
  Category: Word; ID: DWord);
var
  P: Pointer;
begin
  P := PChar(Message);
  if FEventLog = 0 then
    FEventLog := RegisterEventSource(nil, PChar(FName));
  ReportEvent(FEventLog, EventType, Category, ID, nil, 1, 0, @P, nil);
end;
{$endregion}

type
  TLogger = class(TEventLogger)
  private
    Text: string;
  public
    Module: string;
    ModuleVersion: string;
  public
    constructor Create;
    destructor Destroy; override;
    procedure AddModule;
  end;

constructor TLogger.Create;
begin
  inherited Create('Guard')
end;

destructor TLogger.Destroy;
begin
  if Text <> '' then begin
    Text := 'Modified ' + Text;
    LogMessage(Text, EVENTLOG_INFORMATION_TYPE, 0, LOGMSG_INFO);
  end;
  inherited Destroy
end;

procedure TLogger.AddModule;
begin
  if Text <> '' then
    Text := Text + ', ';
  Text := Text + Module + ' ' + ModuleVersion
end;

var
  Log: TLogger;

function CharToHex(Ch:Char): Byte;
begin
  Ch := UpCase(Ch);
  case Ch of
  '0'..'9':
    Result := Ord(Ch) - Ord('0');
  'A'..'F':
    Result := Ord(Ch) - Ord('A') + 10;
  else
    Result := 0
  end
end;

type
  TVersionNumber = array [0..3] of Word;

procedure GetModuleVersion(const ModuleName: PChar; out VersionNumber: TVersionNumber);
var
  ImageBase: Cardinal;
  fname: array[1..MAX_PATH] of char;

  VersionInfo: Pointer;
  VersionInfoSize: Cardinal;

  FileInfo: PVSFixedFileInfo;

  Dummy: Cardinal;
  DummyLen: Cardinal;
begin
  FillChar(VersionNumber, SizeOf(VersionNumber), 0);
  ImageBase := GetModuleHandle(ModuleName);
  if ImageBase = 0 then exit;
  GetModuleFileName(ImageBase, @fname, SizeOf(fname));
  VersionInfoSize := GetFileVersionInfoSize(@fname[1], Dummy);

  if VersionInfoSize > 0 then begin
    GetMem(VersionInfo, VersionInfoSize);
    if GetFileVersionInfo(@fname[1], Dummy, VersionInfoSize, VersionInfo) then begin
      VerQueryValue(VersionInfo, '\', Pointer(FileInfo), DummyLen);
      VersionNumber[0] := HiWord(FileInfo^.dwFileVersionMS);
      VersionNumber[1] := LoWord(FileInfo^.dwFileVersionMS);
      VersionNumber[2] := HiWord(FileInfo^.dwFileVersionLS);
      VersionNumber[3] := LoWord(FileInfo^.dwFileVersionLS);
    end;
    FreeMem(VersionInfo);
  end;
end;

function OpenModuleKey(RootKey: HKEY; const ModuleName: PChar): HKEY;

  { returns number of version components that match.
    0..no match or mismatch
    1..one component exactly, others with wildcard
    2..two components exactly, others with wildcard
    3..three components exactly, other with wildcard
    4..four components exactly }
  function CompareVersion(const VersionNumber: TVersionNumber; const Mask: string): Integer;
  var
    VersionMask: array [Low(TVersionNumber)..High(TVersionNumber)] of string;

    function ParseVersionMask(const Mask: string): Boolean;
    var
      I: Integer;
      TokenBuffer: string;
      Token: string;

      procedure CutToken;
      var
        I: Integer;
      begin
        I := Pos('.', TokenBuffer);
        if I = 0 then begin
          Token := TokenBuffer;
          TokenBuffer := ''
        end else begin
          Token := Copy(TokenBuffer, 1, I - 1);
          Delete(TokenBuffer, 1, I)
        end
      end;

    begin
      TokenBuffer := Mask;

      for I := Low(VersionMask) to High(VersionMask) do begin
        CutToken;
        if Token = '' then
          VersionMask[I] := '*'
        else
          VersionMask[I] := Token;
      end;
      if TokenBuffer = '' then
        Result := True
      else
        Result := False
    end;

    function I2S(I: Word): string;
    begin
      Str(I, Result)
    end;

  var
    I: Integer;
  begin
    Result := 0;
    if not ParseVersionMask(Mask) then exit;
    for I := Low(VersionNumber) to High(VersionNumber) do begin
      if VersionMask[I] = I2S(VersionNumber[I]) then
        Inc(Result)
      else if VersionMask[I] = '*' then
        { it is ok to conntinue, but score won't get higher }
      else begin
        Result := 0;
        exit
      end
    end;
  end;

var
  VersionNumber: TVersionNumber;
  ValueIndex: Cardinal;
  KeyName: array [0..MAX_PATH] of Char;

  CandidateKeyName: string;
  CandidateMatchIndex: Integer;
  MatchIndex: Integer;
begin
  Result := 0;
  GetModuleVersion(ModuleName, VersionNumber);

  CandidateKeyName := '';
  CandidateMatchIndex := 0;

  { we will look at each subkey and of all available choose the one that best
    matches the version number }
  ValueIndex := 0;
  while True do begin
    if RegEnumKey(
      RootKey,
      ValueIndex,
      @KeyName[0],
      SizeOf(KeyName)) <> ERROR_SUCCESS then break;

    { Now KeyName should be the version number match mask in one of following
      formats:
      *
      V1.*
      V1.V2.*
      V1.V2.V3.*
      V1.V2.V3.V4
    }
    MatchIndex := CompareVersion(VersionNumber, PChar(@KeyName[0]));
    if MatchIndex > CandidateMatchIndex then begin
      CandidateMatchIndex := MatchIndex;
      CandidateKeyName := PChar(@KeyName[0])
    end;

    Inc(ValueIndex)
  end;

  if CandidateMatchIndex > 0 then begin
    Log.ModuleVersion := CandidateKeyName;
    if RegOpenKeyEx(
      RootKey,
      PChar(CandidateKeyName),
      0,
      KEY_READ,
      Result) <> ERROR_SUCCESS then
      Result := 0
  end
end;

function SameStr(Str1: PChar; Str2: PChar): Boolean; overload;
var
  I: Integer;
begin
  Result := False;
  if Length(Str1) <> Length(Str2) then exit;
  for I := 0 to Length(Str1) - 1 do
    if UpCase(Str1[I]) <> UpCase(Str2[I]) then exit;
  Result := True;
end;

function SameStr(const Str1: ShortString; const Str2: ShortString): Boolean; overload;
var
  I: Integer;
begin
  Result := False;
  if Length(Str1) <> Length(Str2) then exit;
  for I := 0 to Length(Str1) - 1 do
    if UpCase(Str1[I]) <> UpCase(Str2[I]) then exit;
  Result := True;
end;

type
  TKeyType = (
    kt_Unknown,
    kt_ImageOffset,
    kt_Option_Debug
  );

function GetKeyType(const KeyName: PChar): TKeyType;
var
  CouldBeOffset: Boolean;
  CountZeros: Boolean;
  NumberOfLeadingZeros: Integer;
  I: Integer;
begin
  Result := kt_Unknown;
  if SameStr(KeyName, 'Debug') then begin
    Result := kt_Option_Debug;
    exit
  end;
  CouldBeOffset := False;
  CountZeros := True;
  NumberOfLeadingZeros := 0;
  for I := 0 to Length(KeyName) - 1 do begin
    CouldBeOffset := True;
    if not (UpCase(KeyName[I]) in ['0'..'9', 'A'..'F']) then begin
      CouldBeOffset := False;
      break
    end;
    if KeyName[I] = '0' then begin
      if CountZeros then Inc(NumberOfLeadingZeros)
    end else
      CountZeros := False
  end;
  if CouldBeOffset and (Length(KeyName) - NumberOfLeadingZeros <= 8) then begin
    Result := kt_ImageOffset;
    exit
  end;
end;

function GetRegStrValue(ValueType: Cardinal; var Value; ValueSize: Cardinal): string;
begin
  Result := '';
  case ValueType of
  REG_SZ, REG_EXPAND_SZ:
    begin
      SetLength(Result, ValueSize);
      Move(Value, Result[1], Length(Result));
      while (Result <> '') and (Result[Length(Result)] = #0) do
        SetLength(Result, Length(Result) - 1)
    end;

  else
    // TODO: handle REG_MULTI_SZ
  end;
end;

function GetRegIntValue(ValueType: Cardinal; var Value; ValueSize: Cardinal): Integer;
var
  ErrorCode: Integer;
begin
  case ValueType of
  REG_DWORD:
    Result := PDWORD(@Value)^;

  else
    Val(GetRegStrValue(ValueType, Value, ValueSize), Result, ErrorCode);
    if ErrorCode <> 0 then
      // TODO:
      Result := 0
  end;
end;

function ModuleHooked(RootKey: HKEY; const ModuleName: PChar): Boolean;
var
  ValueIndex: Cardinal;
  ValueName: array [0..MAX_PATH] of Char;
  ValueNameSize: Cardinal;
  KeyName: array [0..MAX_PATH] of Char;
  ValueType: Cardinal;
  Value: array [0..1024 - 1 { TUNE: }] of Byte;
  ValueSize: Cardinal;
  I: Integer;
  ImageOffset: Cardinal;
  Key: HKEY;
  SubKey: HKEY;
  Modified: Boolean;
begin
  Result := False;
  Key := OpenModuleKey(RootKey, ModuleName);
  if Key = 0 then exit;

  Modified := False;

  ValueIndex := 0;

  while True do begin
    ValueNameSize := SizeOf(ValueName);
    ValueSize := SizeOf(Value);
    if RegEnumValue(
      Key,
      ValueIndex,
      ValueName,
      ValueNameSize,
      nil,
      @ValueType,
      @Value[0],
      @ValueSize) <> ERROR_SUCCESS then break;

    case GetKeyType(@ValueName[0]) of
    kt_ImageOffset:
      begin
        { now ValueName is the Hex offset, ValueSize is number of bytes to Poke,
          Value is the array of bytes to poke }
        ImageOffset := 0;
        I := 0;
        while ValueName[I] <> #0 do begin
          ImageOffset := (ImageOffset shl 4) or CharToHex(ValueName[I]);
          Inc(I)
        end;

        case ValueType of
        REG_BINARY, REG_DWORD:
          begin
            PokeModuleByteArray(ModuleName, ImageOffset, @Value[0], ValueSize);
            Modified := True
          end;

        else
          // TODO: convert oter formats, e.g. string representation
        end
      end;

    kt_Option_Debug:
      begin
        if GetRegIntValue(ValueType, Value, ValueSize) <> 0 then
          // keep DLL in memory to enable debugging
          Result := True
      end;

    else
      // TODO: this some unrecognized configuration value. perhaps we should
      // rollback changes applied so far
    end;

    Inc(ValueIndex)
  end;

  if Modified or (ModuleName = nil) then
    Log.AddModule;

  // enum and patch nested DLL modules
  ValueIndex := 0;
  while True do begin
    if RegEnumKey(
      Key,
      ValueIndex,
      @KeyName[0],
      SizeOf(KeyName)) <> ERROR_SUCCESS then break;

    { Now KeyName should be name of the subkey - DLL to be patched }
    if RegOpenKeyEx(
      Key,
      @KeyName[0],
      0,
      KEY_READ,
      SubKey) = ERROR_SUCCESS then begin
      Log.Module := KeyName;
      if ModuleHooked(SubKey, @KeyName[0]) then
        Result := True;
      RegCloseKey(SubKey)
    end;

    Inc(ValueIndex)
  end;

  RegCloseKey(Key);
end;

{ returns True if DLL should stay in memory, False if it can be unloaded }
function HostHooked: Boolean;
var
  fname: array[1..MAX_PATH] of char;
  i: Integer;
  buf: ShortString;
  AppRoot: HKEY;
begin
  Result := False;
  i := GetModuleFileName(0, @fname, sizeof(fname));

  buf := '';
  while fname[i] <> '\' do
  begin
    buf := fname[i] + buf;
    dec(i);
  end;

  Log := TLogger.Create;
  Log.Module := buf;

  buf := 'Software\Microsoft\Windows NT\CurrentVersion\Guard\' + buf + #0;

  AppRoot := 0;
  if RegOpenKeyEx(
    HKEY_CURRENT_USER,
    @buf[1],
    0,
    KEY_READ,
    AppRoot) = ERROR_SUCCESS then begin
    Result := ModuleHooked(AppRoot, nil);
    RegCloseKey(AppRoot);
  end;

  Log.Free;
end;

procedure Leave;
begin
end;

{$region 'This code runs with minimal Run-Time-Library support'}
var
  EntryPoint: Pointer;
  OldEntryCode: array [0..5 - 1] of Byte;

procedure DoEnter; assembler;
begin
  { this code must restore original entrypoint call and set it as next
    instruction. entry point we are restoring has the

      procedure WinMain(instance, previnstance, cmdline, cmdshow); stdcall;

    signature. stdcall calling convention pushes arguments on the stack so
    we don't need to preserve registers because their values are not defined.
    we must preserve the stack only.
    altough this is a 'procedure', we were called through 'jmp' instruction so
    current stack level is the one to preserve }

  { restore original entry point }
  PokeByteArray(EntryPoint, @OldEntryCode[0], SizeOf(OldEntryCode));

  { at this point calling any function from any used DLL is safe
    because all used DLLs were already initialized }
  if HostHooked then begin
    { host was hooked, DLL must stay in memory }
    asm
      jmp EntryPoint
    end
  end else begin
    { we have done the check-and-patch work. No need to stay in memory }

    { we cannot call FreeLibrary(hInstance) directly, because it might
      unmap code section with the code we are currently executing.

      we will setup the stack and jump to FreeLibrary. Upon return the
      FreeLibrary will pop original entrypoint as the return address }
    asm
      push hInstance      { set FreeLibrary argument - our DLL handle }
      push EntryPoint     { set return address for FreeLibrary call }
      jmp FreeLibrary     { FreeLibrary-and-return-to-EntryPoint }
    end
  end
end;

{ Returns True if hooking of current host was enabled }
function HostHookEnabled: Boolean;
var
  fname: array[1..MAX_PATH] of char;
  i: Integer;
  buf: ShortString;
  ConfigFile: TextFile;
  ConfigFileName: ShortString;
  Line: ShortString;
begin
  Result := False;

  i := GetModuleFileName(0, @fname, sizeof(fname));

  buf := '';
  while fname[i] <> '\' do
  begin
    buf := fname[i] + buf;
    dec(i);
  end;

  { now <buf> is the module name }
  i := GetModuleFileName(hInstance, @fname, sizeof(fname));
  if i = 0 then exit;

  { change file extension from '.dll' to '.cfg' }
  fname[i - 2] := 'i';
  fname[i - 1] := 'n';
  fname[i - 0] := 'i';
  SetLength(ConfigFileName, i);
  Move(fname, ConfigFileName[1], i);

  AssignFile(ConfigFile, ConfigFileName);
  {$I-}
  { TODO: use some light-weight reading }
  { TODO: set FileMode to read-only }
  Reset(ConfigFile);
  if IOResult = 0 then begin
    while not Eof(ConfigFile) do begin
      ReadLn(ConfigFile, Line);
      if IOResult <> 0 then break;
      if SameStr(Line, buf) then begin
        Result := True;
        break
      end
    end
  end;
  CloseFile(ConfigFile)
end;

function Enter: Boolean;
var
  img: PImageNTHeaders;
  hModule: Cardinal;
begin
  Result := False;
  hModule := GetModuleHandle(nil);
  if hModule = 0 then exit;
  img := PImageNTHeaders(Cardinal(PImageDosHeader(hModule)^.e_lfanew) + hModule);

  { check if we should care about the host }
  if img^.OptionalHeader.Subsystem = image_Subsystem_Windows_GUI then begin
    if img^.FileHeader.Machine = image_File_Machine_i386 then begin
      Result := True
    end
  end;

  if Result then begin
    if not HostHookEnabled then begin
      { for hosts that are not explicitly enabled don't even
        attempt to hook them as the entry-point redirection may
        cause unpredictable side effects with hosts that do
        that themselves }
      Result := False;
      exit
    end
  end;

  if Result then begin
    EntryPoint := Pointer(hModule + Cardinal(img^.OptionalHeader.AddressOfEntryPoint));

    { we can either use only functions Kernel32.dll or delay hooking.

      we will use delayed hooking. the delaying is implemented by temporarily
      re-routing host's WinMain() to DoEnter.
      DoEnter then restores original WinMain() startup and
      calls the hooker }
    PeekByteArray(EntryPoint, @OldEntryCode[0], SizeOf(OldEntryCode));
    PokeByte(EntryPoint, $E9 { jmp +imm32 }, 1, True);
    PokePointer(Pointer(Cardinal(EntryPoint) + 1), Pointer(Cardinal(@DoEnter) - (Cardinal(EntryPoint) + 5 { size of jmp instruction })), True);
  end
end;
{$endregion}

end.
