program PEDiff;

{$APPTYPE CONSOLE}

uses
  Windows, ExeHdr, Classes, SysUtils, fileSys;

type
  TPatchEntry = record
    FileOffset: Cardinal;
    Value: Byte;
  end;
  TPatchList = class(TList)
  private
    function GetItem(Index: Integer): TPatchEntry;
  public
    procedure Add(FileOffset: Cardinal; Value: Byte);
    property Items[Index: Integer]: TPatchEntry read GetItem; default;
  end;

function TPatchList.GetItem(Index: Integer): TPatchEntry;
var
  Code: Cardinal;
begin
  Code := Cardinal(inherited Get(Index));
  Result.FileOffset := Code shr 8;
  Result.Value := Code and $FF
end;

procedure TPatchList.Add(FileOffset: Cardinal; Value: Byte);
var
  Code: Cardinal;
begin
  Assert(FileOffset <= $FFFFFF);
  Code := (FileOffset shl 8) or Value;
  inherited Add(Pointer(Code))
end;

function GetFileVersion(FileName: string): string;
var
  VersionInfo: Pointer;
  VersionInfoSize: Cardinal;
  FileInfo: PVSFixedFileInfo;

  Dummy: Cardinal;
  DummyLen: Cardinal;

  V1, V2, V3, V4: Word;

  function I2S(I: Word): string;
  begin
    Str(I, Result)
  end;

begin
  V1 := 0;
  V2 := 0;
  V3 := 0;
  V4 := 0;
  VersionInfoSize := GetFileVersionInfoSize(PChar(FileName), Dummy);
  if VersionInfoSize > 0 then begin
    GetMem(VersionInfo, VersionInfoSize);
    if GetFileVersionInfo(PChar(FileName), Dummy, VersionInfoSize, VersionInfo) then begin
      VerQueryValue(VersionInfo, '\', Pointer(FileInfo), DummyLen);
      V1 := HiWord(FileInfo^.dwFileVersionMS);
      V2 := LoWord(FileInfo^.dwFileVersionMS);
      V3 := HiWord(FileInfo^.dwFileVersionLS);
      V4 := LoWord(FileInfo^.dwFileVersionLS);
    end;
    FreeMem(VersionInfo);
  end;
  Result := I2S(V1) + '.' + I2S(V2) + '.' + I2S(V3) + '.' + I2S(V4)
end;

function GetOffsetComment(ImageBase: Cardinal; FileOffset: Cardinal; out ContainsData: Boolean): string;
var
  img: PImageNTHeaders;
  sect: PImageSectionHeader;

  function FindSectionFor(FileOffset: Cardinal): PImageSectionHeader;
  var
    i: integer;
  begin
    Result := PImageSectionHeader(@PChar(img)[sizeof(img^.Signature) + sizeof(img^.FileHeader) + img^.FileHeader.SizeOfOptionalHeader]);
    for i := 1 to img^.FileHeader.NumberOfSections do
      if (Result^.PointerToRawData <> 0) and
         (FileOffset >= Cardinal(Result^.PointerToRawData)) and
         (FileOffset < Cardinal(Result^.PointerToRawData) + Cardinal(Result^.SizeOfRawData)) then
        exit
      else
        Result := PImageSectionHeader(@PChar(Result)[sizeof(TImageSectionHeader)]);
    Result := nil
  end;

var
  SectionName: string;
  I: Integer;
begin
  Result := '';
  ContainsData := True;
  if ImageBase = 0 then exit;

  { module is loaded at address ImageBase }
  img := PImageNTHeaders(Cardinal(PImageDosHeader(ImageBase)^.e_lfanew) + ImageBase);

  { Now we need to find into which PE section the FileOffset belongs }
  sect := FindSectionFor(FileOffset);
  if sect = nil then
    { this some strange situation where FileOffset has no know memory mapping.
      possibly wrong FileOffset was supplied }
    Result := ''
  else begin
    SetLength(SectionName, SizeOf(sect^.Name) + 1);
    FillChar(SectionName[1], 0, Length(SectionName));
    Move(sect^.Name, SectionName[1], SizeOf(sect^.Name));
    I := Pos(#0, SectionName);
    if I <> 0 then
      SetLength(SectionName, I - 1);

    if (sect^.Characteristics and image_Scn_Cnt_Code) <> 0 then
      ContainsData := False;
      
    Result := Format(
      '%.8x=Sect(%s).File(%.8x)+%.8x -> ImageBase(%.8x)+Sect(%s).RVA(%.8x)+%.8x=%.8x', [
      FileOffset,
      SectionName,
      sect^.PointerToRawData,
      FileOffset - Cardinal(sect^.PointerToRawData),
      img^.OptionalHeader.ImageBase,
      SectionName,
      sect^.VirtualAddress,
      FileOffset - Cardinal(sect^.PointerToRawData),
      img^.OptionalHeader.ImageBase + sect^.VirtualAddress + (FileOffset - Cardinal(sect^.PointerToRawData))
    ]);
  end
end;

var
  SrcFile, DstFile: TStream;
  Patches: TPatchList;
  I: Integer;
  SrcByte, DstByte: Byte;
  First, Last: Integer;
  FileOffset: Cardinal;
  ChunkOffset: Cardinal;
  Chunk: string;
  OriginalChunk: string;
  OutputWidth: Integer;
  OutputNeedsSeparator: Boolean;
  LibHandle: Cardinal;
  OffsetComment: string;
  ContainsData: Boolean;
const
  GapSize = 5;  { TUNE: how many unchanged bytes may be included in one
                  continuous patch region (chunk) }
begin
  if ParamCount <> 2 then begin
    WriteLn('Usage: PEDiff orgExe patchedExe');
    exit
  end;
  Patches := TPatchList.Create;
  try
    SrcFile := OSFileSystem.GetFileStream(ParamStr(1), fmOpenRead or fmShareDenyNone);
    try
      DstFile := OSFileSystem.GetFileStream(ParamStr(2), fmOpenRead or fmShareDenyNone);
      try
        if SrcFile.Size <> DstFile.Size then begin
          WriteLn('File sizes differ');
          exit
        end;
        for I := 0 to SrcFile.Size - 1 do begin
          SrcFile.ReadBuffer(SrcByte, SizeOf(SrcByte));
          DstFile.ReadBuffer(DstByte, SizeOf(DstByte));
          if SrcByte <> DstByte then
            Patches.Add(I, DstByte);
        end;

        if Patches.Count = 0 then exit;

        WriteLn('[HKEY_CURRENT_USER\Software\Microsoft\Windows NT\CurrentVersion\Guard\' + ExtractFileName(ParamStr(1)) + '\' + GetFileVersion(ParamStr(1)) + ']');

        LibHandle := LoadLibraryEx(PChar(ParamStr(1)), 0, DONT_RESOLVE_DLL_REFERENCES);
        try
          First := 0;
          while First < Patches.Count do begin
            { now find how many following patches can be included into one chunk.
              let's say chunk is continuous sequence of bytes with not more than
              'GapSize' spaces }
            Last := First;
            while True do begin
              if Last + 1 = Patches.Count then
                { the very last patch }
                break;
              if Patches[Last + 1].FileOffset - Patches[Last].FileOffset <= 1 + GapSize then
                Inc(Last)
              else
                { chunk would be too fragmented }
                break;
            end;

            { gather bytes from Patches[First]..Patches[Last] as one chunk }
            Chunk := Chr(Patches[First].Value);
            ChunkOffset := Patches[First].FileOffset;
            FileOffset := Patches[First].FileOffset + 1;
            Inc(First);
            while First <= Last do begin
              { add not-changed spaces }
              while FileOffset < Patches[First].FileOffset do begin
                DstFile.Position := FileOffset;
                DstFile.ReadBuffer(DstByte, SizeOf(DstByte));
                Chunk := Chunk + Chr(DstByte);
                Inc(FileOffset);
              end;
              Chunk := Chunk + Chr(Patches[First].Value);
              FileOffset := Patches[First].FileOffset + 1;
              Inc(First)
            end;

            SetLength(OriginalChunk, Length(Chunk));
            SrcFile.Position := ChunkOffset;
            SrcFile.ReadBuffer(OriginalChunk[1], Length(OriginalChunk));

            { emit chunk }
            OffsetComment := GetOffsetComment(LibHandle, ChunkOffset, ContainsData);
            if OffsetComment <> '' then begin
              WriteLn('; ', OffsetComment);
            end;

            Write('; original hex:');
            OutputWidth := 15;
            OutputNeedsSeparator := False;
            for I := 1 to Length(OriginalChunk) do begin
              if OutputNeedsSeparator then begin
                Write(',');
                Inc(OutputWidth);
              end;
              if OutputWidth >= 79 then begin
                Write('\');
                WriteLn;
                Write('; ');
                OutputWidth := 2;
              end;
              if OriginalChunk[I] = Chunk[I] then
                Write('..')
              else
                Write(LowerCase(Format('%.2x', [Ord(OriginalChunk[I])])));
              Inc(OutputWidth, 2);
              OutputNeedsSeparator := True
            end;
            WriteLn;

            if ContainsData then begin
              Write('; original asc:');
              OutputWidth := 15;
              OutputNeedsSeparator := False;
              for I := 1 to Length(OriginalChunk) do begin
                if OutputNeedsSeparator then begin
                  Write(',');
                  Inc(OutputWidth);
                end;
                if OutputWidth >= 79 then begin
                  Write('\');
                  WriteLn;
                  Write('; ');
                  OutputWidth := 2;
                end;
                if (OriginalChunk[I] = Chunk[I]) or (not (OriginalChunk[I] in [Chr(33)..Chr(127)])) then
                  Write('..')
                else
                  Write(' ' + OriginalChunk[I]);
                Inc(OutputWidth, 2);
                OutputNeedsSeparator := True
              end;
              WriteLn;
            end;  

            Write(Format('"%.8x"=hex:', [ChunkOffset]));
            OutputWidth := 15;
            OutputNeedsSeparator := False;
            for I := 1 to Length(Chunk) do begin
              if OutputNeedsSeparator then begin
                Write(',');
                Inc(OutputWidth);
              end;
              if OutputWidth >= 79 then begin
                Write('\');
                WriteLn;
                Write('  ');
                OutputWidth := 2;
              end;
              Write(LowerCase(Format('%.2x', [Ord(Chunk[I])])));
              Inc(OutputWidth, 2);
              OutputNeedsSeparator := True
            end;
            WriteLn;

            WriteLn;

            First := Last + 1
          end;
        finally
          FreeLibrary(LibHandle);
        end
      finally
        DstFile.Free
      end
    finally
      SrcFile.Free
    end;
  finally
    Patches.Free
  end
end.
