unit Tools;

interface

procedure PokeByte(Address: Pointer; const Value: Byte; Count: Integer = 1; IsCodePointer: Boolean = False);
procedure PokePointer(Address: Pointer; const Value: Pointer; IsCodePointer: Boolean = False);
procedure PokeBytes(Address: Pointer; const Values: array of Byte; IsCodePointer: Boolean = False);
procedure PokeByteArray(Address: Pointer; Values: PByte; Count: Integer; IsCodePointer: Boolean = False);
procedure PeekByteArray(Address: Pointer; Values: PByte; Count: Integer);

procedure PokeModuleByte(const ModuleName: PChar; FileOffset: Cardinal; const Value: Byte; const Count: Integer = 1);
procedure PokeModuleBytes(const ModuleName: PChar; FileOffset: Cardinal; const Values: array of Byte);
procedure PokeModuleByteArray(const ModuleName: PChar; FileOffset: Cardinal; Values: PByte; Count: Integer);

implementation
uses
  Windows, ExeHdr;

function UnlockMem(Address: Pointer; Count: Integer; var state: integer): Pointer;
begin
  Result := Address;
  VirtualProtect(Address, Count, PAGE_EXECUTE_WRITECOPY, @state)
end;

procedure LockMem(Address: Pointer; Count: Integer; state: integer);
var
  dummy: integer;
begin
  VirtualProtect(Address, Count, state, @dummy)
end;

procedure PokeByte(Address: Pointer; const Value: Byte; Count: Integer; IsCodePointer: Boolean);
var
  lck: integer;
  p: Pointer;
begin
  if Address <> nil then begin
    p := UnlockMem(Address, SizeOf(Byte) * Count, lck);
    FillChar(p^, Count, Value);
    LockMem(Address, SizeOf(Byte) * Count, lck);
    if IsCodePointer then
      FlushInstructionCache(GetCurrentProcess, Address, SizeOf(Byte) * Count)
  end
end;

procedure PokePointer(Address: Pointer; const Value: Pointer; IsCodePointer: Boolean);
begin
  PokeByteArray(Address, System.PByte(@Value), SizeOf(Value), IsCodePointer)
end;

procedure PokeBytes(Address: Pointer; const Values: array of Byte; IsCodePointer: Boolean);
var
  I: Integer;
begin
  if Address <> nil then begin
    for I := Low(Values) to High(Values) do begin
      PokeByte(Address, Values[I], 1, IsCodePointer);
      Inc(Cardinal(Address))
    end
  end
end;

procedure PokeByteArray(Address: Pointer; Values: System.PByte; Count: Integer; IsCodePointer: Boolean);
begin
  if Address <> nil then begin
    while Count > 0 do begin
      PokeByte(Address, Values^, 1, IsCodePointer);
      Inc(Cardinal(Address));
      Inc(Cardinal(Values));
      Dec(Count)
    end
  end
end;

procedure PeekByteArray(Address: Pointer; Values: System.PByte; Count: Integer);
begin
  if Address <> nil then begin
    while Count > 0 do begin
      Values^ := PByte(Address)^;
      Inc(Cardinal(Address));
      Inc(Cardinal(Values));
      Dec(Count)
    end
  end
end;

function FileOffsetToImagePointer(const ModuleName: PChar; FileOffset: Cardinal; out IsCodePointer: Boolean): Pointer;
var
  ImageBase: Cardinal;
  img: PImageNTHeaders;
  sect: PImageSectionHeader;

  function FindSectionFor(FileOffset: Cardinal): PImageSectionHeader;
  var
    i: integer;
  begin
    Result := PImageSectionHeader(@PChar(img)[sizeof(img^.Signature) + sizeof(img^.FileHeader) + img^.FileHeader.SizeOfOptionalHeader]);
    for i := 1 to img^.FileHeader.NumberOfSections do
      if (Result^.PointerToRawData <> 0) and
         (FileOffset >= Cardinal(Result^.PointerToRawData)) and
         (FileOffset < Cardinal(Result^.PointerToRawData) + Cardinal(Result^.SizeOfRawData)) then
        exit
      else
        Result := PImageSectionHeader(@PChar(Result)[sizeof(TImageSectionHeader)]);
    Result := nil
  end;

begin
  IsCodePointer := False;
  ImageBase := GetModuleHandle(ModuleName);
  if ImageBase <> 0 then begin
    { module is loaded at address ImageBase }
    img := PImageNTHeaders(Cardinal(PImageDosHeader(ImageBase)^.e_lfanew) + ImageBase);

    { Now we need to find into which PE section the FileOffset belongs.
      Then we will know where was the section mapped in memory (relative to the ImageBase)
      and thus we can calculate where was the FileOffset mapped to }
    sect := FindSectionFor(FileOffset);
    if sect = nil then
      { this some strange situation where FileOffset has no know memory mapping.
        possibly wrong FileOffset was supplied }
      Result := nil
    else begin
      Result := Pointer(ImageBase + Cardinal(sect^.VirtualAddress) + (FileOffset - Cardinal(sect^.PointerToRawData)));
      if (sect^.Characteristics and image_Scn_Cnt_Code) <> 0 then
        IsCodePointer := True
    end
  end else
    Result := nil
end;

procedure PokeModuleByte(const ModuleName: PChar; FileOffset: Cardinal; const Value: Byte; const Count: Integer);
var
  Address: Pointer;
  IsCodePointer: Boolean;
begin
  Address := FileOffsetToImagePointer(ModuleName, FileOffset, IsCodePointer);
  PokeByte(Address, Value, Count, IsCodePointer)
end;

procedure PokeModuleBytes(const ModuleName: PChar; FileOffset: Cardinal; const Values: array of Byte);
var
  Address: Pointer;
  IsCodePointer: Boolean;
begin
  Address := FileOffsetToImagePointer(ModuleName, FileOffset, IsCodePointer);
  PokeBytes(Address, Values, IsCodePointer)
end;

procedure PokeModuleByteArray(const ModuleName: PChar; FileOffset: Cardinal; Values: System.PByte; Count: Integer);
var
  Address: Pointer;
  IsCodePointer: Boolean;
begin
  Address := FileOffsetToImagePointer(ModuleName, FileOffset, IsCodePointer);
  PokeByteArray(Address, Values, Count, IsCodePointer)
end;

end.
